\ fontcursor.4th
\
\ Copyright 1987, 1998  The Open Group
\
\ Permission to use, copy, modify, distribute, and sell this software and its
\ documentation for any purpose is hereby granted without fee, provided that
\ the above copyright notice appear in all copies and that both that
\ copyright notice and this permission notice appear in supporting
\ documentation.
\

154  constant  XC_num_glyphs 
  0  constant  XC_X_cursor 
  2  constant  XC_arrow 
  4  constant  XC_based_arrow_down 
  6  constant  XC_based_arrow_up 
  8  constant  XC_boat 
 10  constant  XC_bogosity 
 12  constant  XC_bottom_left_corner 
 14  constant  XC_bottom_right_corner 
 16  constant  XC_bottom_side 
 18  constant  XC_bottom_tee 
 20  constant  XC_box_spiral 
 22  constant  XC_center_ptr 
 24  constant  XC_circle 
 26  constant  XC_clock 
 28  constant  XC_coffee_mug 
 30  constant  XC_cross 
 32  constant  XC_cross_reverse 
 34  constant  XC_crosshair 
 36  constant  XC_diamond_cross 
 38  constant  XC_dot 
 40  constant  XC_dotbox 
 42  constant  XC_double_arrow 
 44  constant  XC_draft_large 
 46  constant  XC_draft_small 
 48  constant  XC_draped_box 
 50  constant  XC_exchange 
 52  constant  XC_fleur 
 54  constant  XC_gobbler 
 56  constant  XC_gumby 
 58  constant  XC_hand1 
 60  constant  XC_hand2 
 62  constant  XC_heart 
 64  constant  XC_icon 
 66  constant  XC_iron_cross 
 68  constant  XC_left_ptr 
 70  constant  XC_left_side 
 72  constant  XC_left_tee 
 74  constant  XC_leftbutton 
 76  constant  XC_ll_angle 
 78  constant  XC_lr_angle 
 80  constant  XC_man 
 82  constant  XC_middlebutton 
 84  constant  XC_mouse 
 86  constant  XC_pencil 
 88  constant  XC_pirate 
 90  constant  XC_plus 
 92  constant  XC_question_arrow 
 94  constant  XC_right_ptr 
 96  constant  XC_right_side 
 98  constant  XC_right_tee 
100  constant  XC_rightbutton 
102  constant  XC_rtl_logo 
104  constant  XC_sailboat 
106  constant  XC_sb_down_arrow 
108  constant  XC_sb_h_double_arrow 
110  constant  XC_sb_left_arrow 
112  constant  XC_sb_right_arrow 
114  constant  XC_sb_up_arrow 
116  constant  XC_sb_v_double_arrow 
118  constant  XC_shuttle 
120  constant  XC_sizing 
122  constant  XC_spider 
124  constant  XC_spraycan 
126  constant  XC_star 
128  constant  XC_target 
130  constant  XC_tcross 
132  constant  XC_top_left_arrow 
134  constant  XC_top_left_corner 
136  constant  XC_top_right_corner 
138  constant  XC_top_side 
140  constant  XC_top_tee 
142  constant  XC_trek 
144  constant  XC_ul_angle 
146  constant  XC_umbrella 
148  constant  XC_ur_angle 
150  constant  XC_watch 
152  constant  XC_xterm 



